/**
 * DiCloud - GENI project
 * 
 * (C) 2010 - University of Massachusetts Amherst
 * 
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package edu.umass.cs.geni.dicloud.server;

import java.io.File;
import java.io.FileInputStream;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Properties;

import org.apache.log4j.Logger;

import edu.umass.cs.geni.dicloud.accounting.AWSAccounting;
import edu.umass.cs.geni.dicloud.handlers.AWSPlugin;
import edu.umass.cs.geni.dicloud.handlers.EBSHandler;
import edu.umass.cs.geni.dicloud.handlers.EC2Handler;
import edu.umass.cs.geni.dicloud.handlers.S3Handler;

/**
 * Starts the AWS accounting service (including its embedded Derby database).
 * EC2 and S3 handlers are instantiated and the S3 proxy is started.
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class DiCloudServer extends UnicastRemoteObject implements
    DiCloudServerInterface
{

  private static final long  serialVersionUID      = -3032210214045048346L;
  public static final String RMI_NAME_REGISTRATION = "dicloud/server";
  private static Logger      logger                = Logger.getRootLogger();
  private boolean            quit;
  private Object             sleepObject           = new Object();

  /**
   * 
   * @throws RemoteException
   */
  public DiCloudServer() throws RemoteException
  {
	quit = false;
  }

  /**
   * @param args
   *          1. optional RMI port number
   */
  public static void main(String[] args)
  {
	if (args.length < 1)
	{
	  System.out
		  .println("Usage: java -cp ... edu.umass.cs.geni.dicloud.server.DiCloudServer dicloud.properties");
	  System.exit(1);
	}
	DiCloudServer server;
	try
	{
	  server = new DiCloudServer();
	  server.startServer(args[0]);
	}
	catch (RemoteException e)
	{
	  // TODO Auto-generated catch block
	  e.printStackTrace();
	}
  }

  private void startServer(String propFile)
  {
	Properties diCloudProperties = new Properties();
	AWSPlugin plugin = null;
	EC2Handler ec2Handler = null;
	AWSAccounting awsAccounting = null;
	try
	{
	  logger.info("Reading properties from " + propFile);
	  diCloudProperties.load(new FileInputStream(new File(propFile)));

	  // Start the RMI registry
	  String rmiPortString = diCloudProperties.getProperty("rmi.port.number");
	  int rmiPort = 1099;
	  if (rmiPortString != null)
		rmiPort = Integer.valueOf(rmiPortString);
	  logger.info("Starting RMI registry on port " + rmiPort);
	  java.rmi.registry.LocateRegistry.createRegistry(rmiPort);

	  logger.info("Registering DiCloud server");
	  Naming.rebind(RMI_NAME_REGISTRATION, this);

	  // Create the AWS accounting service
	  plugin = new AWSPlugin();
	  plugin.setLogger(Logger.getRootLogger());
	  plugin.initialize();

	  awsAccounting = plugin.getAwsAccouting();
	  String minimumCreditString = diCloudProperties
		  .getProperty("aws.minimum.credit");
	  if (minimumCreditString != null)
		awsAccounting.setMinimumCredit(Double.valueOf(minimumCreditString));
	  else
		awsAccounting.setMinimumCredit(0);
	  logger.info("Registering AWS accouting service");
	  Naming.rebind(AWSAccounting.RMI_NAME_REGISTRATION, awsAccounting);

	  logger.info("Registering EC2 handler");
	  ec2Handler = new EC2Handler(plugin);
	  ec2Handler.initialize();
	  EC2Server ec2Server = new EC2Server(plugin, ec2Handler);
	  Naming.rebind(EC2Server.RMI_NAME_REGISTRATION, ec2Server);

	  logger.info("Registering S3 handler");
	  S3Handler s3Handler = new S3Handler(plugin);
	  s3Handler.initialize();
	  S3Server s3Server = new S3Server(plugin, s3Handler);
	  Naming.rebind(S3Server.RMI_NAME_REGISTRATION, s3Server);

	  logger.info("Registering EBS handler");
	  EBSHandler ebsHandler = new EBSHandler(plugin);
	  ebsHandler.initialize();
	  EBSServer ebsServer = new EBSServer(plugin, ebsHandler);
	  Naming.rebind(EBSServer.RMI_NAME_REGISTRATION, ebsServer);

	  logger.info("Waiting for console connections");
	  while (!quit)
	  {
		synchronized (sleepObject)
		{
		  sleepObject.wait();
		}
	  }

	}
	catch (Exception e)
	{
	  // TODO Auto-generated catch block
	  e.printStackTrace();
	}
	finally
	{
	  awsAccounting.shutdown();
	}

  }

  public void shutdown() throws RemoteException
  {
	logger
	    .info("Received a shutdown request. Terminating accounting services...");
	quit = true;
	synchronized (sleepObject)
	{
	  sleepObject.notifyAll();
	}
  }

}
