/**
 * DiCloud - GENI project
 * 
 * (C) 2010 - University of Massachusetts Amherst
 * 
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package edu.umass.cs.geni.dicloud.server;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Properties;

/**
 * Interface to start/stop/probe EC2 instances.
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public interface EC2Interface extends Remote
{
  /**
   * Create a new EC2 instance by invoking ec2-run-instances.
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EC2_AMI_ID: AMI id to use to create instances
   *  - AmazonProperties.EC2_INSTANCE_TYPE: type of the instance to create
   *  - AmazonProperties.EC2_INSTANCE_NUMBER: number of instances to create
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   *  - AmazonProperties.EC2_AVAILABILITY_ZONE: availability zone where to create the instance
   *  - AmazonProperties.EC2_KEY_PAIR_NAME: registered key pair name to use (one will be generated 
   *  automatically and returned in AmazonProperties.EC2_KEY_PAIR if none is provided)
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EC2_INSTANCE_ID: Instance id
   * AmazonProperties.EC2_KEY_PAIR_NAME: Name of the key pair that has been registered
   * AmazonProperties.EC2_KEY_PAIR: Key pair created for this instance
   * </pre>
   */
  public abstract Properties startInstance(Properties p) throws Exception,
	  RemoteException;

  /**
   * Stop an EC2 instance by invoking ec2-terminate-instances
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EC2_INSTANCE_ID: AMI instance id to probe
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   *  - AmazonProperties.EC2_KEY_PAIR_NAME: Key pair name used for this instance
   *  (if the keypair is a temporary keypair starting with KEY_PAIR_PREFIX, it is deleted)
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * </pre>
   */
  public abstract Properties stopInstance(Properties p) throws Exception,
	  RemoteException;

  /**
   * Invoke ec2-describe-instances to probe the instance. The expected
   * properties are:
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.EC2_AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.EC2_AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EC2_INSTANCE_ID: AMI instance id to probe
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: AMI status (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EC2_INSTANCE_ID: Instance id
   * AmazonProperties.EC2_AMI_ID: AMI id
   * AmazonProperties.EC2_PUBLIC_NAME: Public name
   * AmazonProperties.EC2_PRIVATE_NAME: Private name
   * AmazonProperties.EC2_PUBLIC_IP: Public IP
   * AmazonProperties.EC2_PRIVATE_IP: Private IP
   * AmazonProperties.EC2_INSTANCE_STATE: Instance state
   * AmazonProperties.EC2_INSTANCE_TYPE: Instance type
   * AmazonProperties.EC2_KEY_PAIR_NAME: Name of the key pair used
   * AmazonProperties.EC2_START_TIME: Start time
   * AmazonProperties.EC2_AVAILABILITY_ZONE: Availability zone
   * </pre>
   */
  public abstract Properties describeInstance(Properties p) throws Exception,
	  RemoteException;

}
